from .Field import Field
from typing import Any
import random
from danielutils import validate


class RationalField(Field):
    __int_float = [[int, float], None, None]

    @validate(None, int, int)
    def __init__(self, degree: int = 1, modulu: int = 1):
        """Constructor for RationalField class

        Args:
            degree (int, optional): the degree of the field. Defaults to 1.
            modulu (int, optional): the modulu of the field. Defaults to 1.
        """
        super().__init__("Q", 0, 1, degree, modulu)

    @validate(None, __int_float, __int_float)
    def random(self, min: int = -10, max: int = 10) -> float:
        """returns a random element of the field

        Args:
            min (int, optional): the minimum value for an element. Defaults to -10.
            max (int, optional): the maximum value for an element. Defaults to 10.

        Raises:
            ValueError: if min >= max

        Returns:
            float: a random element of the field
        """
        if min >= max:
            raise ValueError("'min' cant be >= 'max'")
        f = random.randint
        sign = 1 if f(0, 1) == 1 else -1
        nominator = f(min, max)
        denominator = f(min, max)
        while (denominator == 0):
            denominator = f(min, max)
        return sign*nominator/denominator

    @validate(None, None, bool)
    def __contains__(self, value: Any, quiet: bool = False) -> bool:
        """checks if the value is in the field

        Args:
            value (Union[int, float]): the value to check
            quiet (bool, optional): whether to supress warning. Defaults to False.

        Returns:
            bool: True if value is in the field, False otherwise
        """
        if not quiet:
            print(
                f"due to a finite amount of values, every number is rational so take note")
            print("to turn these messages of off, use __contains__(value, quiet=True)")
        if isinstance(value, [int, float]):
            return True
        return False



from .Calculable import Calculable
from .PolynomialSimple import PolynomialSimple
from .PolynomialFraction import PolynomialFraction
from .PolynomialCompact import PolynomialCompact
from .Polynomial import Polynomial
# from .Expression import Expression



import random
from .Field import Field
from typing import Any, Union
from danielutils import isoneof, validate


class RealField(Field):
    __int_float = [[int, float], None, None]

    @validate(None, int, int)
    def __init__(self, degree: int = 1, modulu: int = 1):
        """Constructor for RealField class

        Args:
            degree (int, optional): the degree of the field. Defaults to 1.
            modulu (int, optional): the modulu of the field. Defaults to 1.
        """
        super().__init__("R", 0, 1, degree, modulu)

    @validate(None, __int_float, __int_float)
    def random(self, min: Union[int, float] = -10, max: Union[int, float] = 10) -> float:
        """returns a random element of the field

        Args:
            min (int, optional): the minimum value for an element. Defaults to -10.
            max (int, optional): the maximum value for an element. Defaults to 10.

        Returns:
            float: _description_
        """
        return random.uniform(min, max)

    def __contains__(self, value: Any) -> bool:
        """checks if the value is in the field

        Args:
            value (Any): the value to check

        Returns:
            bool: True if value is in the field, False otherwise
        """
        if isoneof(value, [int, float]):
            return True
        return False



from typing import Callable


def message(msg: str) -> Callable:
    def wrapper(func):
        def inner(*args, **kwargs):
            print(msg)
            return func(*args, **kwargs)
        return inner
    return wrapper


def depracated(extra_info: str = "") -> Callable:
    def wrapper(func):
        def inner(*args, **kwargs):
            print(
                f'{func.__module__}.{func.__name__}(...) is depracated!', extra_info, sep="\n")
            return func(*args, **kwargs)
        return inner
    return wrapper



from __future__ import annotations
from functools import reduce
from ..utils import almost_equal, areinstances
from typing import Union, Any, Sequence
from .Field import Field, RealField
from .Complex import Complex
from danielutils import validate, isoneof
from ..BaseClasses import Vector____


class Vector__(Vector____):
    @validate(None, Sequence, Field)
    def __init__(self, values: list[Any], field: Field = None) -> None:
        """Initialize a vector with the given values and field

        Args:
            values (list[Any]): the values of the vector
            field (Field, optional): if None Defaults to RealField(len(values)).
        """
        self.__values = values
        # TODO add default field detection
        self.field = RealField(len(values)) if not field else field

    @validate(None, bool)
    def __str__(self, raw: bool = False) -> str:
        """Return a string representation of the vector

        Args:
            raw (bool, optional): wheter to print raw data ot to pretty print it. Defaults to False.

        Returns:
            str: a string representation of the vector
        """
        if raw:
            result = "["
            for v in self:
                result += str(v) + ", "
            return result[:-2]+"]"
        max_individual_length = reduce(lambda sub_res, value: max(
            sub_res, len(str(value))), self.__values, 0)
        res = ""
        hs = "|"+"-"*(max_individual_length+2)+"|\n"
        res += hs
        for v in self:
            res += "|"+str(v).center(max_individual_length+2)+"|\n"
            res += hs
        return res

    @validate(None, Vector____)
    def __add__(self, other: Vector) -> Vector:
        """Add two vectors together

        Args:
            other (Vector): the vector to add to self

        Raises:
            TypeError: if other is not a vector
            ValueError: if the vectors have different fields
            ValueError: if the vectors have different lengths

        Returns:
            Vector: the sum of the two vectors
        """
        if self.field != other.field:
            raise ValueError("Vectors must have the same field")
        if len(self.__values) != len(other.__values):
            raise ValueError("Vectors must have the same length")
        return Vector([self[i] + other[i] for i in range(len(self))], self.field)

    @validate(None, Vector____)
    def __radd__(self, other: Vector) -> Vector:
        """Add a vector to a number

        Args:
            other (Vector): the vector to add to self

        Raises:
            whatever __add__ raises

        Returns:
            Vector: the sum of the two vectors
        """
        try:
            return self.__add__(other)
        except Exception as e:
            raise e

    @validate(None, Vector____)
    def __sub__(self, other: Vector) -> Vector:
        """Subtract two vectors

        Args:
            other (Vector): the vector to subtract from self

        Raises:
            TypeError: if other is not a vector
            ValueError: if the vectors have different fields
            ValueError: if the vectors have different lengths

        Returns:
            Vector: the difference of the two vectors
        """
        if self.field != other.field:
            raise ValueError("Vectors must have the same field")
        if len(self.__values) != len(other.__values):
            raise ValueError("Vectors must have the same length")
        return Vector([self.__values[i] - other.__values[i] for i in range(len(self.__values))], self.field)

    @validate(None, Vector____)
    def __rsub__(self, other: Vector) -> Vector:
        """Subtract a vector from a number

        Args:
            other (Vector): the vector to subtract from self

        Raises:
            whatever __neg__ or __add__ raises

        Returns:
            Vector: the difference of the two vectors
        """
        try:
            return (-self) + other
        except Exception as e:
            raise e

    def __neg__(self) -> Vector:
        """Negate a vector
            x => -x
        Returns:
            Vector: the negated vector
        """
        return Vector([-self.__values[i] for i in range(len(self.__values))], self.field)

    @validate(None, [[int, float, Complex], None, None])
    def __mul__(self, num: Union[int, float, Complex]) -> Vector:
        """ Multiply a vector by a number or a complex number

        Args:
            num (Union[int, float, Complex]): the number to multiply the vector by

        Raises:
            TypeError: if num is not a number

        Returns:
            Vector: the multiplied vector
        """
        return Vector([num * self.__values[i] for i in range(len(self.__values))], self.field)

    @validate(None, [[int, float, Complex], None, None])
    def __rmul__(self, num: Union[int, float, Complex]) -> Vector:
        """Multiply a vector by a number

        Args:
            num (float): the number to multiply the vector by

        Raises:
            whatever __mul__ raises

        Returns:
            Vector: the product of the vector and the number
        """
        try:
            return self.__mul__(num)
        except Exception as e:
            raise e

    @validate(None, [[int, float, Complex], None, None])
    def __truediv__(self, other: Union[int, float, Complex]) -> Vector:
        """Divide a vector by a number or a complex number

        Args:
            other (Union[int, float, Complex]): the number to divide the vector by

        Raises:
            Exception: whatever __mul__ raises
            ValueError: if other is a vector, vector/vector is not defined

        Returns:
            Vector: the divided vector
        """
        try:
            if isoneof(other, [int, float, Complex]):
                return self.__mul__(1/other)
        except Exception as e:
            raise e
        raise ValueError("cant divide by a vector")

    def __rtruediv__(self, _):
        """Divide a vector by a number or a complex number

        Raises:
            ValueError: */vector is not defined
        """
        raise ValueError("cant divide by a vector")

    @validate(None, int)
    def __getitem__(self, index: int) -> Any:
        """Get the value at a given index

        Args:
            index (int): the index of the value to get

        Raises:
            TypeError: if index is not an int
            IndexError: if index is out of range

        Returns:
            Any: the value at the given index
        """
        if not (0 <= index < len(self)):
            raise IndexError("Vector index out of range")
        return self.__values[index]

    @validate(None, int, None)
    def __setitem__(self, index: int, value: Any) -> None:
        """Set the value at a given index

        Args:
            index (int): the index of the value to set
            value (Any): the value to set at the given index

        Raises:
            TypeError: if index is not an int
            ValueError: if the value is not in the field
        """
        if not (0 <= index < len(self)):
            raise ValueError("index out of range")
        # FIXME validate tha value is valid
        self.__values[index] = value

    def __iter__(self):
        """Iterate over the values in the vector

        Returns:
            list iterator for all the elements in the vector
        """
        return iter(self.__values)

    @validate(None, None, bool, bool)
    def __eq__(self, other: Vector, use_almost_equal: bool = True, check_field_equality: bool = True) -> bool:
        """Check if two vectors are equal

        Args:
            other (Vector): the vector to compare to self
            use_almost_equal (bool, optional): wheter to use 'almost_equal' functions. Defaults to True.
            check_field_equality (bool, optional): wheter to check if the fields are identical. Defaults to True.

        Raises:
            TypeError: if use_almost_equal and check_field_equality are not booleans

        Returns:
            bool: True if the vectors are equal, False otherwise
        """
        if not isinstance(other, Vector):
            return False
        if check_field_equality:
            if self.field != other.field:
                return False
        if len(self) != len(other):
            return False

        def compare_func(a, b): return a == b
        if use_almost_equal:
            compare_func = almost_equal

        for i in range(len(self)):
            if not compare_func(self[i], other[i]):
                return False
        return True

    def __ne__(self, other: Vector, use_almost_equal: bool = True, check_field_equality: bool = True) -> bool:
        """Check if two vectors are not equal

        Args:
            other (Vector): the vector to compare to self
            use_almost_equal (bool, optional): wheter to use 'almost_equal' functions. Defaults to True.
            check_field_equality (bool, optional): wheter to check if the fields are identical. Defaults to True.

        Raises:
            raises the same as __eq__

        Returns:
            bool: True if the vectors are not equal, False otherwise
        """
        try:
            return not self.__eq__(other, use_almost_equal, check_field_equality)
        except Exception as e:
            raise e

    def __len__(self) -> int:
        """Get the length of the vector

        Returns:
            int: the length of the vector
        """
        return len(self.__values)

    def __hash__(self) -> int:
        """Get the hash of the vector

        Returns:
            int: the hash of the vector
        """
        return hash((v for v in self))


class Vector(Vector__):

    @staticmethod
    def random(min: Any = -10, max: Any = 10, size: int = 10,  def_value: Any = None, f: Field = RealField()) -> Vector:
        """Generate a random vector with random values in the given range.

        Args:
            min (Any, optional): the minimum value for each element. Defaults to -10.
            max (Any, optional): the maximum value for each element. Defaults to 10.
            size (int, optional): the size of the vector. Defaults to 10.
            def_value (Any, optional): Default value to be put in all of the elements. Defaults to None.
            f (Field, optional): the field from which to genetare elements. Defaults to RealField().

        Raises:
            TypeError: size must be an integer
            ValueError: size must be greater than 0

        Returns:
            Vector: a random vector
        """
        if not isinstance(size, int):
            raise TypeError("size must be an integer")
        if not (0 < size):
            raise ValueError("size must be greater than 0")
        return Vector([f.random(min, max) if def_value is None else def_value for _ in range(size)])

    @staticmethod
    @validate(int, Field, None)
    def from_size(size: int, field: Field, default_value: Any = None) -> Vector:
        """Create a vector of the given size with the given field and default value

        Args:
            size (int): the size of the vector
            field (Field): the field of the vector
            default_value (Any, optional): the value to intialize the vector with. if is None will default to field._zero.

        Returns:
            Vector: a vector of the given size with the given field and default value
        """
        if default_value is None:
            default_value = field.zero
        return Vector([default_value for _ in range(size)], field.__class__(size, field.modulu))

    @ staticmethod
    @validate(int, int, Field)
    def e(i: int, size: int, field: Field) -> Vector:
        """Create the ith element of the e vector of the given size and field

        Args:
            i (int): the index in which to put field.one
            size (int): the size of the vector
            field (Field): the field of the vector

        Returns:
            Vector: the ith element of the e vector of the given size and field
        """
        v = Vector.from_size(size, field, field.zero)
        v[i] = field.one
        return v

    @ property
    def conjugate(self) -> Vector:
        """Return the conjugate of the vector

        Returns:
            Vector: the conjugate of the vector
        """
        arr = []
        for v in self:
            if getattr(v, "conjugate", None):
                if callable(getattr(v, "conjugate")):
                    arr.append(v.conjugate())
                else:
                    arr.append(v.conjugate)
            else:
                arr.append(v)
        return Vector(arr)

    def norm(self) -> float:
        """Get the norm of the vector

        Returns:
            float: the norm of the vector
        """
        return sum([x ** 2 for x in self]) ** 0.5

    @validate(None, Vector__)
    def dot(self, other: Vector) -> Vector:
        """Get the dot product of two vectors

        Args:
            other (Vector): the vector to dot with self

        Raises:
            TypeError: if other is not a vector
            ValueError: if the vectors are not in the same field
            ValueError: if the vectors are not the same length

        Returns:
            Vector: the dot product of the vectors
        """
        if not isinstance(other, Vector):
            raise TypeError("Vector can only be multiplied by another Vector")
        if self.field != other.field:
            raise ValueError("Vectors must have the same field")
        if len(self.__values) != len(other.__values):
            raise ValueError("Vectors must have the same length")
        return Vector([self.__values[i] * other.__values[i] for i in range(len(self.__values))])

    def normalize(self) -> Vector:
        """Get the normalized vector

        Returns:
            Vector: the normalized vector
        """
        return self / self.norm()

    def projection_onto(self, value) -> Vector:
        """return the projection of self onto value which can be another vector or a Span

        Args:
            value (Union[Vector, Span]): the value to project onto

        Raises:
            TypeError: if value is not a vector or a span
            ValueError: if the vectors are not in the same field
            ValueError: if the vectors are not the same length

        Returns:
            Vector: the projection of self onto value
        """
        from .Span import Span
        if not isoneof(value, [Vector, Span]):
            raise TypeError("v must be of type Vector or Span")
        is_span = isinstance(value, Span)
        if not is_span and len(value) != len(self):
            raise ValueError("value must have the same length as self")
        if is_span and len(value[0]) != len(self):
            raise ValueError(
                "the span's vectors must have the same length as self")
        from ..la2.InnerProduct import StandardInnerProduct as sip
        if not is_span:
            return sip(self, value)/sip(value, value)*value
        else:
            return value.projection_of(self)

    def duplicate(self) -> Vector:
        """Get a duplicate of the vector

        Returns:
            Vector: a duplicate of the vector
        """
        return Vector([v for v in self], self.field)



from .Field import Field
from ..Complex import Complex
import random
from typing import Any


class ComplexField(Field):
    # validated at super().__init__
    def __init__(self, degree: int = 1, modulu: int = 1) -> None:
        """_summary_

        Args:
            degree (int, optional): the degree of the field. Defaults to 1.
            modulu (int, optional): the modulu of the field. Defaults to 1.
        """
        super().__init__("C", Complex(0, 0),
                         Complex(1, 0), degree, modulu)

    def random(self, min: int = -10, max: int = 10) -> Complex:
        """_summary_

        Args:
            min (int, optional): the minimum value for an element. Defaults to -10.
            max (int, optional): the maximum value for an element. Defaults to 10.

        Returns:
            Complex: _description_
        """
        return Complex.random(min, max, random.uniform)

    def __contains__(self, value: Any) -> bool:
        """checks if the value is in the field

        Args:
            value (Any): the value to check

        Returns:
            bool: True if value is in the field, False otherwise
        """
        if isinstance(value, [int, float, Complex]):
            return True
        return False



from typing import Any


def newton_raphson(polynomial, x, derivative=None, iterations: int = 1) -> Any:
    der = polynomial.derivative if not derivative else derivative
    for _ in range(iterations):
        x = x - polynomial(x) / der(x)
    return x



from __future__ import annotations
from typing import Any, Callable
from ...utils import are_operators_implemnted, almost_equal, areinstances
from danielutils import validate, abstractmethod

from ...BaseClasses import Field____


class Field(Field____):
    """An interface class to create derived classes
    """
    @staticmethod
    @validate(Field____, int)
    def is_field(field: Field, iterations: int = 100) -> bool:
        """will check if all of the field axioms hold, returns a probablistic answer!

        Args:
            field (Field): the field to check
            iterations (int, optional): how many iterations to perform Defaults to 100.

        Raises:
            NotImplementedError: will rise if a neccessary method is not implemented

        Returns:
            bool: True if all of the field axioms hold
        """
        if not are_operators_implemnted(type(field.random())):
            raise NotImplementedError(
                "One of the nescesary operators for calculation was not implemented")

        def checker(var_count: int, rule: Callable[[], bool], exclude=None) -> bool:
            if exclude is None:
                exclude = []
            for _ in range(iterations):
                vars = [field.random() for _ in range(var_count)]
                for v in vars:
                    if v in exclude:
                        break
                else:
                    if not rule(*vars):
                        return False
            return True

        def associativity() -> bool:
            def addition() -> bool:
                return checker(3, lambda a, b, c: almost_equal((a+b)+c, a+(b+c)))

            def multiplication() -> bool:
                return checker(3, lambda a, b, c: almost_equal((a*b)*c, a*(b*c)))
            return all([addition(), multiplication()])

        def commutativity() -> bool:
            def addition() -> bool:
                return checker(2, lambda a, b: almost_equal(a+b, b+a))

            def multiplication() -> bool:
                return checker(2, lambda a, b: almost_equal(a*b, b*a))
            return all([addition(), multiplication()])

        def distributivity() -> bool:
            def addition() -> bool:
                return checker(3, lambda a, b, c: almost_equal(a*(b+c), a * b + a * c))

            def multiplication() -> bool:
                return checker(3, lambda a, b, c: almost_equal((a+b)*c, a * c + b * c))
            return all([addition(), multiplication()])

        def identity() -> bool:
            def addition() -> bool:
                return checker(1, lambda a: almost_equal(a+field.zero, a, field.zero+a))

            def multiplication() -> bool:
                return checker(1, lambda a: almost_equal(a*field.one, a, field.one*a))
            return all([addition(), multiplication()])

        def inverses() -> bool:
            def addition() -> bool:
                return checker(1, lambda a: almost_equal(a + (-a), field.zero, (-a)+a))

            def multiplication() -> bool:
                return checker(1, lambda a: almost_equal(a * (1/a), field.one, (1/a)*a), [0])
            return all([addition(), multiplication()])

        for func in [associativity, commutativity, distributivity, identity, inverses]:
            if not func():
                return False
        return True

    @validate(None, None, None, None, int, int, bool)
    def __init__(self, name: Any, zero: Any, one: Any, degree: int = 1, modulu: int = 1, validate: bool = False) -> None:
        """Constructor for Field class

        Args:
            name (Any): the name of the field
            zero (Any): the zero element of the field
            one (Any): the one element of the field
            degree (int, optional): the degree of the field. Defaults to 1.
            modulu (int, optional): the modulu of the field. Defaults to 1.
            validate (bool, optional): whether to call is_field on construction to validate. Defaults to False.

        Raises:
            TypeError: if degree or modulu is not an integer
            ValueError: if validate=True and is_field return false
        """
        if not areinstances([degree, modulu], int):
            raise TypeError("'degree' and 'modulu' must be of type 'int'")
        self.name = name
        self.modulu = modulu
        self.degree = degree
        self.zero = zero
        self.one = one
        if validate:
            if not Field.is_field(self):
                raise ValueError(
                    "This is not a field as one or more of the axioms do not check-out")

    def __str__(self) -> str:
        """returns the string representation of the field

        Returns:
            str: the string representation of the field
        """
        return f"{self.name}{self.degree}%{self.modulu}"

    @validate(None, Field____)
    def __eq__(self, other: Field) -> bool:
        return self.name == other.name and self.modulu == other.modulu and self.degree == other.degree and self.zero == other.zero and self.one == other.one

    @abstractmethod
    def random(self, min: Any = -10, max: Any = 10) -> Any:
        """returns a random element of the field

        Args:
            min (Any, optional): the minimum value for an element. Defaults to -10.
            max (Any, optional): the maximum value for an element. Defaults to 10.

        Raises:
            NotImplementedError: will rise if this method is not implemented in a derived class as this is an abstract method

        Returns:
            Any: a random element of the field
        """
        pass

    @abstractmethod
    def __contains__(self, value: Any, *args, **kwargs) -> bool:
        """checks if a value is in the field

        Args:
            value (Any): the value to check

        Raises:
            NotImplementedError: will rise if this method is not implemented in a derived class as this is an abstract method

        Returns:
            bool: True if the value is in the field
        """
        pass



from typing import Tuple, Any

bracket_pairs = {
    "(": ")",
    "[": "]",
    "{": "}",
    ")": "(",
    "]": "[",
    "}": "{"
}


def validate_brackets(s: str) -> bool:
    if len(s) == 0:
        return True
    brackets = ["(", ")", "[", "]", "{", "}"]
    stack = []
    for c in s:
        if c in brackets:
            if c in ["(", "[", "{"]:
                stack.append(c)
            else:
                if len(stack) == 0:
                    return False
                if stack[-1] != brackets[brackets.index(c)-1]:
                    return False
                stack.pop()
    return len(stack) == 0


open_brackets = ["(", "[", "{"]
close_brackets = [")", "]", "}"]


def bracketify(input) -> list[Tuple[str, int]]:
    res = []
    for i, c in enumerate(input):
        if c in open_brackets+close_brackets:
            res.append((c, i))
    return res


def insert_multiplication(input: str) -> str:
    brackets = bracketify(input)
    stack = []
    # recent = None
    change = 0
    for i in range(len(brackets)-1):
        if brackets[i][0] in open_brackets:
            stack.append(brackets[i])
        elif brackets[i][0] in close_brackets:
            if stack[-1][0] == bracket_pairs[brackets[i][0]]:
                # recent = stack.pop()
                if brackets[i+1][0] in open_brackets and brackets[i+1][1] == brackets[i][1] + 1:
                    input = f"{input[:brackets[i][1]+1+change]}*{input[brackets[i][1]+1+change:]}"
                    change += 1
    return input


def open_power(input: str) -> str:
    stack = []
    i = 0
    change = 0
    while i < len(input)-1:
        c = input[i]
        if c in open_brackets:
            stack.append((c, i))
        else:
            if c in close_brackets:
                if stack[-1][0] == bracket_pairs[c]:
                    recent = stack.pop()
                    if input[i+1] == "^":
                        try:
                            # extract power end index
                            j = i+2
                            while j < len(input):
                                c2 = input[j]
                                if not c2.isdigit():
                                    break
                                j += 1
                            # parse power
                            power = int(input[i+2:j])
                            # amount of digits in the power
                            size = len(str(power))
                            # the text to insert 'power' times
                            text_to_insert = input[recent[1]:i+1]

                            # insertion
                            change = 0
                            for _ in range(power-1):
                                input = f"{input[:i+1+change]}{text_to_insert}{input[i+change+2+size:]}"
                                change += len(text_to_insert)
                            i += change
                        except Exception as e:
                            pass

        i += 1
    return input


def split_not_between_brackets(input: str, symbols: list[str]) -> list[str]:
    input = insert_multiplication(input)
    indecies = []
    stack = []
    split_indecies = []
    open_i = 0
    for i, c in enumerate(input):
        if c in symbols:
            split_indecies.append(i)
        if c in open_brackets:
            stack.append(c)
            open_i = i
        elif c in close_brackets:
            if stack[-1] == bracket_pairs[c]:
                stack.pop()
            if len(stack) == 0:
                indecies.append((open_i, i))
    # remove indecies that are inside brackets
    i = 0
    while i < len(split_indecies):
        index = split_indecies[i]
        for j in indecies:
            if j[0] < index < j[1]:
                split_indecies.remove(index)
                i -= 1
                break
        i += 1
    # split
    res = []
    order = []
    start = 0
    for end in range(len(input)):
        if end in split_indecies:
            order.append(input[end])
            res.append(input[start:end].strip())
            start = end+1
    res.append(input[start:].strip())
    return res, order


def concat_horizontally(lst: list[Any], sep: str = " ", end: str = "") -> str:
    res = ""
    strs = [str(v)for v in lst]
    prev_char_indecies = [0 for _ in range(len(lst))]
    char_indecies = [0 for _ in range(len(lst))]
    while sum(char_indecies) < sum([len(s) for s in strs]):
        # find indecies for each entry
        to_advance = set([i for i in range(len(lst))])
        to_remove = set()
        while len(to_advance) > 0:
            for i in to_advance:
                if strs[i][char_indecies[i]] == '\n':
                    to_remove.add(i)
                    continue
                char_indecies[i] += 1
            for v in to_remove:
                to_advance.remove(v)
            to_remove.clear()
        # print acordingly
        for vec_index in range(len(strs)):
            res += strs[vec_index][prev_char_indecies[vec_index]:char_indecies[vec_index]]+sep
            prev_char_indecies[vec_index] = char_indecies[vec_index]+1
            char_indecies[vec_index] += 1
        res += "\n"
    return res+end


def split_many(lst: list[str], mark: str) -> list[str]:
    res = []
    for s in lst:
        res.extend(s.split(mark))
    return res


def split_if_any(string: str, marks: list[str]) -> Tuple[list[str], list[str]]:
    splits = []
    order = []
    start = 0
    for end, c in enumerate(string):
        if c in marks:
            splits.append(string[start:end].strip())
            order.append(c)
            start = end+1
    splits.append(string[start:].strip())
    return splits, order



from typing import Callable
from functools import wraps


def memo(func: Callable) -> Callable:
    d = dict()

    @wraps(func)
    def wrapper(*args, **kwargs):
        tup = (*args, *list(kwargs.keys()), *list(kwargs.values()))
        if tup in d:
            return d[tup]
        res = func(*args, **kwargs)
        d[tup] = res
        return res
    return wrapper



from .string_manipulation import bracket_pairs


def string_splitter(lst: list[str], symbol: str, excluders: list[str] = list(bracket_pairs.keys())) -> list[str]:
    res = []
    for string in lst:
        res.extend(string.strip().split(symbol))
    return res


def stacker(string: str, lst: list[str]) -> list[str]:
    res = []
    for c in string:
        if c in lst:
            res.append(c)
    return res


def sign(op):
    if op is '+':
        return 1
    return -1



from __future__ import annotations
from typing import Any, Union, Sequence
import enum
import copy
import functools
from typing import Tuple
from ..utils import almost_equal, areinstances, check_foreach
from .Complex import Complex
from .Vector import Vector
from .Field import Field, RealField
from .Span import Span
from danielutils import validate, isoneof, NotImplemented
from ..BaseClasses import Matrix____


class MatrixOperationType(enum.Enum):
    ROW_SWITCHING = "row_switching"
    ROW_MULTIPLICATION = "row_multiplication"
    ROW_ADDITION = "row_addition"
    COL_SWITCHING = "column_switching"
    COL_MULTIPLICATION = "column_multiplication"
    COL_ADDITION = "column_addition"


MOT = MatrixOperationType


class Matrix__(Matrix____):
    @validate(None, Sequence, Field)
    def __init__(self, mat: list[list[Any]], field: Field = RealField()) -> None:
        """Create a Matrix

        Args:
            mat (list[list[Any]]): a list of lists of elements (2D array)
            field (Field, optional): the field that the elements of the matrix is over. Defaults to RealField().

        Raises:
            TypeError: will rise if mat is not a list of lists
        """
        if not isinstance(mat, list) or not all([isinstance(v, list) for v in mat]):
            raise TypeError("Matrix must be a 2d array")
        self._matrix = mat
        self._rows = len(mat)
        self._cols = len(mat[0])
        self.field = field

    @validate(None, int)
    def __getitem__(self, index: int) -> list[Any]:
        """wil return the row at the given index

        Args:
            index (int): the index of desire row

        Raises:
            TypeError: will rise if index is not an int
            ValueError: will rise if index is out of range

        Returns:
            list[Any]: the row at the given index
        """
        if not (0 <= index < self._rows):
            raise ValueError("Index out of range")
        return self._matrix[index]

    @validate(None, int, list)
    def __setitem__(self, index: int, value: list[Any]) -> None:
        if not 0 <= index < self._rows:
            raise ValueError("Index out of range")
        # FIXME validate all items in list are of same type of this matrix
        self._matrix[index] = value

    @validate(None, int)
    def __str__(self, turnc: int = 2) -> str:
        """will return a string representation of the matrix

        Args:
            turnc (int, optional): how many digist to turnicate. Defaults to 2.

        Returns:
            str: the srtring representation
        """
        def round_if_possible(v):
            if hasattr(v, '__round__'):
                if almost_equal(round(v), v):
                    v = round(v)
            return v

        def turnacate(v):
            v = str(v)
            return v[:v.index(".")+1+turnc] if "." in v else v

        def find_spaceing():
            res = 0
            for row in self:
                for v in row:
                    v = round_if_possible(v)
                    v = turnacate(v)
                    res = max(res, len(str(v)))
            return res
        spacing = find_spaceing()+2
        n = len(self[0])
        vl = "|"
        hl = (1+spacing)*n*"-" + "-" + "\n"
        result = hl
        for i, row in enumerate(self._matrix):
            result += vl
            for v in row:
                v = round_if_possible(v)
                v = turnacate(v)
                result += str(v).center(spacing)+vl
            result += "\n"+hl
        return result

    @validate(None, Matrix____)
    def __add__(self, other: Matrix__) -> Matrix__:
        """will add two matrices together and return the result

        Args:
            other (Matrix): the matrix to add to current one

        Raises:
            TypeError: will rise if other is not a matrix
            ValueError: will rise if the matrices are not of the same size

        Returns:
            Matrix: the result matrix
        """
        if not isinstance(other, Matrix):
            raise TypeError("Matrix can only be added to another Matrix")
        if self._rows != other._rows or self._cols != other._cols:
            raise ValueError("Matrices must have the same dimensions")
        return Matrix([[self._matrix[i][j] + other._matrix[i][j] for j in range(self._cols)]
                       for i in range(self._rows)])

    def __neg__(self) -> Matrix__:
        """will return the negative of the matrix

        Returns:
            Matrix: the negative matrix
        """
        return Matrix([[-self._matrix[i][j] for j in range(self._cols)]
                       for i in range(self._rows)])

    @validate(None, Matrix____)
    def __sub__(self, other: Matrix__) -> Matrix__:
        """will subtract two matrices and return the result

        Args:
            other (Matrix): the matrix to subtract from current one

        Raises:
            TypeError: will rise if other is not a matrix
            ValueError: will rise if the matrices are not of the same size

        Returns:
            Matrix: the result matrix
        """
        if not isinstance(other, Matrix):
            raise TypeError(
                "Matrix can only be subtracted from another Matrix")
        if self._rows != other._rows or self._cols != other._cols:
            raise ValueError("Matrices must have the same dimensions")
        return Matrix([[self._matrix[i][j] - other._matrix[i][j] for j in range(self._cols)]
                       for i in range(self._rows)])

    def __mul__(self, other: Any) -> Union[Matrix__, Vector]:
        """will multiply the matrix with the given value and return the result

        Args:
            other (Union[int, float, Complex, Vector, Matrix, Calculable]): the value to multiply with

        Raises:
            ValueError: will rise if the value is not one of the above types

        Returns:
            Any: [int, float, Complex, Vector, Matrix, Calculable]
        """
        from ..la2 import Calculable
        if not isoneof(other, [int, float, complex, Vector, Matrix, Calculable]):
            raise ValueError(
                "Can only multiply with a int, float, complex, Vector, Matrix or Calculable")
        if isoneof(other, [int, float, Complex, Calculable]):
            res: list[list[Any]] = []
            for i in range(len(self)):
                res.append([])
                for j in range(len(self[0])):
                    res[i].append(self[i][j] * other)
            return Matrix(res)
        if isinstance(other, Vector):
            if self._cols != len(other):
                raise ValueError(
                    "Matrix and Vector must have the same number of rows")
            return Vector([sum([self._matrix[i][j] * other[j] for j in range(self._cols)])
                           for i in range(self._rows)])
        if isinstance(other, Matrix):
            if self._cols != other._rows:
                raise ValueError(
                    "Matrix and Matrix must have matching sizes: self.cols == other.rows")
            return Matrix([[sum([self._matrix[i][j] * other._matrix[j][k] for j in range(self._cols)])
                            for k in range(other._cols)] for i in range(self._rows)])

    def __rmul__(self, other: Union[int, float, Complex, Vector, Matrix__]) -> Union[float, Complex, Vector, Matrix__]:
        """will multiply the matrix with the given value and return the result

        Args:
            other (Union[int, float, Complex, Vector, Matrix, Calculable]): the value to multiply with

        Raises:
            ValueError: will rise if the value is not one of the above types

        Returns:
            Any: [int, float, Complex, Vector, Matrix, Calculable]
        """
        if isoneof(other, [int, float, Complex]):
            return self.__mul__(other)
        if isinstance(other, Vector):
            raise TypeError(
                "Matrix can only be multiplied by a vector from the right")
        if isinstance(other, Matrix):
            if self._cols != other._rows:
                raise ValueError(
                    "Matrix and Matrix must have the same number of columns")
            return Matrix([[sum([self._matrix[i][j] * other._matrix[j][k] for j in range(self._cols)])
                            for k in range(other._cols)] for i in range(self._rows)])
        raise TypeError(
            f"cant perform {type(other)}*Matrix.\ncan only be multiplied by a:\n\tint\n\tfloat\n\tComplex\n\tVector\n\tMatrix]")

    @validate(None, Matrix____, bool)
    def __eq__(self, other: Matrix__, use_almost_equale: bool = True) -> bool:
        """will compare two matrices and return True if they are equal

        Args:
            other (Matrix): the other matrix to compare to
            use_almost_equale (bool, optional): Specifies wheter to use almost_equal when comparing Matricies. Defaults to True.

        Raises:
            TypeError: will rise it the other object is not a Matrix
            TypeError: wiil rise if use_almost_equale is not a bool

        Returns:
            bool: True if the objects are equal, False otherwise
        """
        if not isinstance(other, Matrix):
            raise TypeError(f"cant complare 'Matrix' with '{type(other)}'")
        if not isinstance(use_almost_equale, bool):
            raise TypeError("use_almost_equale must be a boolean")
        if self._rows != other._rows or self._cols != other._cols:
            return False
        if use_almost_equale:
            for i in range(len(self)):
                for j in range(len(self[i])):
                    if not almost_equal(self[i][j], other[i][j]):
                        return False
        if any([self._matrix[i][j] != other._matrix[i][j] for i in range(self._rows) for j in range(self._cols)]):
            return False
        return True

    @validate(None, Matrix____, bool)
    def __ne__(self, other: Matrix__, use_almost_equale: bool = True) -> bool:
        """will compare two matrices and return True if they are not equal

        Args:
            other (Matrix): the other matrix to compare to
            use_almost_equale (bool, optional): Specifies wheter to use almost_equal when comparing Matricies. Defaults to True.

        Raises:
            TypeError: will rise if __eq__ would raise an error

        Returns:
            bool: True if the objects are not equal, False otherwise
        """
        try:
            return not self.__eq__(other, use_almost_equale)
        except TypeError as e:
            raise e

    def __len__(self) -> int:
        """will return the number of rows in the matrix

        Returns:
            int: the number of rows in the matrix
        """
        return self._rows

    @validate(None, int)
    def __pow__(self, value: int) -> Matrix__:
        """will raise the matrix to the given power and return the result in a new matrix

        Args:
            value (int): the power to raise the matrix to

        Raises:
            TypeError: will rise if value is not an int
            ValueError: will rise if value is less than 0

        Returns:
            Matrix: the result of the power
        """
        if isinstance(value, float):
            if int(value) != value:
                raise ValueError("value must be an integer")
            value = int(value)
        if not isinstance(value, int):
            raise TypeError("value must be an int")
        if not (0 <= value):
            raise ValueError("value must be atleast than 0")
        res = self
        for _ in range(value-1):
            res *= self
        return res

    def __iter__(self) -> list:
        return iter(self._matrix)


class Matrix(Matrix__):

    @staticmethod
    @validate(Vector)
    def fromVector(vec: Vector) -> Matrix:
        """Create a Matrix from a single Vector, the matrix will have 1 column and the same number of rows as the vector

        Args:
            vec (Vector): the vector to create a mtrix from

        Raises:
            TypeError: wiil rise if vec is not a Vector

        Returns:
            Matrix: the result
        """
        if not isinstance(vec, Vector):
            raise TypeError("Vector must be a Vector")
        return Matrix([[v] for v in vec], vec.field)

    @staticmethod
    @validate([Sequence, lambda seq: areinstances(seq, Vector), "all elements must be instances of class 'Vector'"])
    def from_vectors(vecs: Sequence[Vector]) -> Matrix:
        """Create a Matrix from a list of Vectors, the matrix will have the same number of columns as the number of vectors and the same number of rows as the number of elements in a vector

        Args:
            vecs (list[Vector]): the list of vector to create a matrix from

        Raises:
            TypeError: will rise if vecs is not a list of Vectors
            ValueError: will rise if vecs are not over the same field

        Returns:
            Matrix: the result
        """
        if not check_foreach(vecs, lambda v: v.field == vecs[0].field):
            raise ValueError("vectors are not over the same field")
        mat: list[list[Any]] = []
        for i in range(len(vecs[0])):
            mat.append([])
            for j in range(len(vecs)):
                mat[i].append(vecs[j][i])
        return Matrix(mat, field=vecs[0].field)

    @staticmethod
    @validate(int, int, Field, float, float)
    def random(rows: int, cols: int, f: Field = RealField(), min: float = -10, max: float = 10, ) -> Matrix:
        """Create a random Matrux acording to params

        Args:
            rows (int): the amount of rows in matrix
            cols (int): the amount of columns in matrix
            f (Field, optional): the field to create the matrix over. Defaults to RealField().
            min (float, optional): the minimum value from the field. Defaults to -10.
            max (float, optional): the maximum value from the field. Defaults to 10.

        Returns:
            Matrix: the result
        """
        return Matrix([[f.random(min, max) for _ in range(cols)]for __ in range(rows)], field=f)

    @staticmethod
    @validate([Sequence, lambda seq: areinstances(seq, Matrix), None])
    def from_jordan_blocks(blocks: Sequence[Matrix]) -> Matrix:
        """Create a Matrix from a list of Jordan blocks

        Args:
            blocks (list[Matrix]): the list of Jordan blocks to create a matrix from

        Returns:
            Matrix: the result
        """
        # group blocks by eigenvectors
        eigenvalues = set([m[0][0] for m in blocks])
        matricies: list[list[Matrix]] = []
        for eigenvalue in eigenvalues:
            matricies.append([])
            for m in blocks:
                if m[0][0] == eigenvalue:
                    matricies[-1].append(m)

        def inside_sorter(b1, b2):
            size1 = len(b1)
            size2 = len(b2)
            return size2-size1
        # sort each sub list for big to small
        for i in range(len(eigenvalues)):
            matricies[i].sort(key=functools.cmp_to_key(inside_sorter))
        # rebuild blocks array
        blocks = []
        for mat_list in matricies:
            blocks.extend(mat_list)
        # calculate final size
        total_size = sum([len(m) for m in blocks])
        # initialize result
        res = Matrix([[0 for __ in range(total_size)] for _ in range(total_size)],
                     field=blocks[0].field)
        # set values inside result
        offset = 0
        for block in blocks:
            for i in range(len(block)):
                for j in range(len(block)):
                    res[i+offset][j+offset] = block[i][j]
            offset += len(block)
        return res

    @staticmethod
    @validate(int, None)
    def create_jordan_blcok(size: int, eigenvalue: Any) -> Matrix:
        """Create a Jordan block from a size and an eigenvalue

        Args:
            size (int): the size of the block
            eigenvalue (Any): the eigenvalue of the block

        Returns:
            Matrix: the result
        """
        m = Matrix.identity(size)-Matrix.identity(size)
        for i in range(size-1):
            m[i+1][i] = 1
        return Matrix.identity(size)*eigenvalue + m

    @staticmethod
    @validate([int, lambda x: x > 0, "must be positive integer"])
    def identity(size: int) -> Matrix:
        """Create identity matrix if given size

        Args:
            size (int): the size of the matrix

        Returns:
            Matrix: the result
        """
        arr = [[0 for __ in range(size)] for _ in range(size)]
        for i in range(size):
            arr[i][i] = 1
        return Matrix(arr)

    @property
    def kernel(self) -> Union[Vector, Span]:
        """Get the kernel of the matrix

        Returns:
            Union[Vector, Span]: the result
        """
        solution = self.solve(Vector.from_size(len(self), self.field))
        if solution is None:
            return Span(Vector([0 for _ in range(len(self))]))
        return solution

    @property
    def image(self) -> list[Vector]:
        """Get the image of the matrix

        Returns:
            list[Vector]: the result
        """
        from .VectorSpace import VectorSpace
        return VectorSpace(self.field).standard_basis() - self.kernel

    @property
    @NotImplemented
    def adjugate(self) -> Matrix:
        pass
        # TODO implement adjugate

    @property
    def row_space(self) -> list[Vector]:
        return [Vector(r) for r in self]

    @property
    def column_space(self) -> list[Vector]:
        res: list[Vector] = []
        for j in range(len(self[0])):
            arr = []
            for i in range(len(self)):
                arr.append(self[i][j])
            res.append(Vector(arr))
        return res

    @property
    def rank(self) -> int:
        """Get the rank of the matrix

        Returns:
            int: the result
        """
        # TODO fix rank calculation this is not correct
        tmp = self.gaussian_elimination()
        rank = 0
        for i in range(min(len(tmp), len(tmp[0]))):
            for v in tmp[i]:
                if v != tmp.field.zero:
                    rank += 1
                    break

        # for i in tmp.__matrix:
        #     if not all([i == 0 for i in i]):
        #         rank += 1
        return rank

    @property
    def determinant(self) -> Any:
        """Calculate the determinant of the matrix

        Raises:
            ValueError: will rise if the matrix is not square

        Returns:
            Any: the result
        """
        if self._rows != self._cols:
            raise ValueError("Matrix must be square")
        if self._rows == 1:
            return self._matrix[0][0]
        if self._rows == 2:
            return self._matrix[0][0] * self._matrix[1][1] - self._matrix[0][1] * self._matrix[1][0]
        return sum([self._matrix[i][0] * ((-1)**i) * self.minor(i, 0) for i in range(self._rows)])

    @property
    def is_invertiable(self) -> bool:
        """returns wheter the matrix is invertible

        Returns:
            bool: True if the matrix is invertible, False otherwise
        """
        return self.determinant != 0

    @property
    def inverse(self) -> Matrix:

        if not self.is_invertiable:
            raise ValueError("Matrix must be invertible")
        _, res = self.gaussian_elimination_with(Matrix.identity(len(self)))
        return res

    @property
    def is_square(self) -> bool:
        """returns wheter the matrix is square

        Returns:
            bool: True if the matrix is square, False otherwise
        """
        return self._rows == self._cols

    @property
    def is_symmetrical(self) -> bool:
        """returns wheter the matrix is symmetrical

        Returns:
            bool: True if the matrix is symmetrical, False otherwise
        """
        if not self.is_square:
            return False
        for i in range(len(self)):
            for j in range(i, len(self)):
                if self[i][j] != self[j][i]:
                    return False
        return True

    @property
    def is_asymmetrical(self) -> bool:
        """returns wheter the matrix is asymmetrical

        Returns:
            bool: True if the matrix is asymmetrical, False otherwise
        """
        if not self.is_square:
            return False
        for i in range(len(self)):
            for j in range(i, len(self)):
                if self[i][j] != -self[j][i]:
                    return False
        return True

    @property
    def conjugate_transpose(self) -> Matrix:
        return self.transpose().conjugate()

    @property
    def is_diagonialable(self) -> bool:
        """returns wheter the matrix is diagonialable

        Returns:
            bool: True if the matrix is diagonialable, False otherwise
        """
        for eigenvalue in set(self.eigenvalues):
            if not (self.algebraic_multiplicity(eigenvalue) == self.geometric_multiplicity(eigenvalue)):
                return False
        return True

    @property
    def is_projection(self) -> bool:
        return self**2 == self

    @property
    def is_nilpotent(self) -> bool:
        # TODO implement nilpotency check
        pass

    @property
    def eigenvalues(self) -> list[Any]:
        """Get the eigenvalues of the matrix

        Returns:
            list: the result
        """
        return self.characteristic_polynomial.roots

    @property
    def jordan_form(self) -> Matrix:
        """Get the Jordan form of the matrix

        Raises:
            ValueError: will rise if the matrix is not square

        Returns:
            Matrix: the result
        """
        if not self.is_square:
            raise ValueError("Matrix must be square")
        Block = Matrix.from_jordan_blocks
        J = Matrix.create_jordan_blcok
        blocks = []
        eigenvalues = self.eigenvalues
        V0 = Vector([0 for _ in range(len(self))])
        In = Matrix.identity(len(self))
        for eigenvalue in set(eigenvalues):
            alg = eigenvalues.count(eigenvalue)
            counts = []
            M_lamda: Matrix = self-eigenvalue*In
            for k in range(1, alg):
                counts.append((M_lamda**k).kernel.dim)
            if len(counts) == 1:
                if counts[0] == 1:
                    blocks.append(J(alg, eigenvalue))
            else:
                diagram = [[0 for _ in range(counts[0])]
                           for __ in range(counts[-1])]
                pass
        return Block(blocks)

    @property
    def chain_basis(self) -> list[Vector]:
        # TODO implement chain basis calculation
        pass

    @property
    def characteristic_polynomial(self):
        """Get the characteristic polynomial of the matrix

        Raises:
            ValueError: will rise if the matrix is not square

        Returns:
            PolynomialSimple: the result
        """
        from ..la2 import PolynomialSimple
        if not self.is_square:
            raise ValueError("Matrix must be square")
        return (Matrix.identity(len(self))*PolynomialSimple([1], [1])-self).determinant

    @property
    def minimal_polynomial(self):
        """Get the minimal polynomial of the matrix

        Returns:
            _type_: the result
        """
        from ..la2 import PolynomialSimple
        res = PolynomialSimple([1], [0])
        eigenvalues = self.eigenvalues
        for eigenvalue in set(eigenvalues):
            alg = eigenvalues.count(eigenvalue)
            tmp_matrix: Matrix = (
                self-eigenvalue*Matrix.identity(len(self)))
            power = 0
            ker = tmp_matrix.kernel
            if isinstance(ker, Span):
                if ker.dim == 1:
                    power = alg
                elif ker.dim == alg:
                    power = 1
                else:
                    # M0 = Matrix.identity(len(self))*0
                    # print(M0)
                    # original = tmp_matrix
                    # power += 1
                    # print(tmp_matrix)
                    # while tmp_matrix != M0:
                    #     tmp_matrix *= original
                    #     print(tmp_matrix)
                    #     power += 1
                    _0 = Vector([0 for _ in range(len(ker[0]))])
                    for v in ker:
                        curr_power = 1
                        u = tmp_matrix*v
                        while u != _0:
                            u = tmp_matrix*v
                            curr_power += 1
                        power = max(power, curr_power)

            res *= PolynomialSimple([1, -eigenvalue], [1, 0])**power
        return res

    @validate(None, int, int)
    def cofactor(self, row_to_ignore: int, col_to_ignore: int) -> Matrix:
        """will return the cofactor of the matrix at the given position

        Args:
            row_to_ignore (int): the row to ignore
            col_to_ignore (int): the column to ignore

        Raises:
            ValueError: will rise if the given row or column is out of range

        Returns:
            Matrix: the cofactor of the matrix at the given position
        """
        if not ((0 <= row_to_ignore < self._rows) and (0 <= col_to_ignore < self._cols)):
            raise ValueError("Row or column index out of range")
        res: list[list[Any]] = []
        for i, row in enumerate(self._matrix):
            if i == row_to_ignore:
                continue
            res.append([])
            for j, col in enumerate(self._matrix[i]):
                if j == col_to_ignore:
                    continue
                res[i if i < row_to_ignore else i -
                    1].append(self._matrix[i][j])
        return Matrix(res)

    @validate(None, int, int)
    def minor(self, row_to_ignore: int, col_to_ignore: int) -> Any:
        """will return the minor of the matrix at the given position (the determinant of the cofactor)

        Args:
            row_to_ignore (int): the row to ignore
            col_to_ignore (int): the column to ignore

        Raises:
            will rise whatever cofator and/or determinant will raise

        Returns:
            Any: the minor of the matrix at the given position
        """
        return self.cofactor(row_to_ignore, col_to_ignore).determinant

    def transpose(self) -> Matrix:
        """will return the transpose of the matrix

        Returns:
            Matrix: the transpose of the matrix
        """
        return Matrix([[self._matrix[j][i] for j in range(self._cols)]
                       for i in range(self._rows)])

    def conjugate(self) -> Matrix:
        """will return the conjugate of the matrix

        Returns:
            Matrix: the conjugate of the matrix
        """
        arr: list[list[Any]] = []
        for row in self:
            arr.append([])
            for v in row:
                if hasattr(v, "conjugate"):
                    if callable(getattr(v, "conjugate")):
                        arr[-1].append(v.conjugate())
                    arr[-1].append(v.conjugate)
                else:
                    arr[-1].append(v)
        return Matrix(arr, self.field)

    def reorgenize_rows(self) -> None:
        """Reorgenizes the rows of the matrix
        """
        def comparer(a: list[float], b: list[float]) -> bool:
            def first_not_zero_index(row: list[float]) -> int:
                i = 0
                while i < len(row):
                    if row[i] != self.field.zero:
                        break
                    i += 1
                return i

            a_index = first_not_zero_index(a)
            b_index = first_not_zero_index(b)
            if a_index == b_index:
                return 0
            return -1 if a_index > b_index else 1

        self._matrix = sorted(
            self._matrix, key=functools.cmp_to_key(comparer), reverse=True)

    def apply_operation(self, operation: MatrixOperationType, iv1, iv2, operate_with=None) -> Matrix:
        if not isinstance(operation, MatrixOperationType):
            raise TypeError("can only apply MatrixOperationType")
        if operation == MOT.ROW_MULTIPLICATION:
            self[iv1] = [iv2*v for v in self[iv1]]
        elif operation == MOT.ROW_ADDITION:
            self[iv1] = [self[iv1][i]+self[iv2][i]
                         for i in range(len(self[iv1]))]
        elif operation == MOT.COL_SWITCHING:
            self[iv1], self[iv2] = self[iv2], self[iv1]
        elif operation == MOT.COL_SWITCHING:
            for i in range(len(self[0])):
                self[i][iv1], self[i][iv2] = self[i][iv2], self[i][iv1]
        elif operation == MOT.COL_MULTIPLICATION:
            for i in range(len(self[0])):
                self[i][iv1] *= iv2
        elif operation == MOT.COL_ADDITION:
            for i in range(len(self[0])):
                self[i][iv1] += self[i][iv2]

        return operate_with

    @validate(None, [[Vector, Matrix__], None, None])
    def concat(self, other: Union[Vector, Matrix]) -> Matrix:
        """will concatenate the matrix with the given matrix and return the result in a new matrix

        Args:
            other (Union[Vector, Matrix]): the matrix/vector to concatenate with

        Raises:
            TypeError: will rise if other is not a Matrix or Vector
            ValueError: will rise if the len(self) != len(other)

        Returns:
            Matrix: the result of the concatenation
        """
        if len(other) != len(self):
            raise ValueError(
                "can only concat with a Matrix|Vector with the same number of rows")
        if isinstance(other, Vector):
            other = Matrix.fromVector(other)
        res = []
        for i in range(len(self)):
            res.append(self[i] + other[i])
        return Matrix(res)

    @validate(None, int)
    def split(self, index: int) -> Tuple[Matrix, Matrix]:
        """will split the matrix into two matrices at the given index

        Args:
            index (int): the index to split the matrix at

        Raises:
            TypeError: will rise if index is not an int
            ValueError: will rise if index is out of range

        Returns:
            Tuple[Matrix, Matrix]: the two matrices resulting from the split
        """
        if not 0 <= index <= len(self[0]):
            raise ValueError("index out of range")
        if index == len(self[0]):
            return self, None
        return Matrix([[self[i][j] for j in range(index)] for i in range(len(self))], self.field), Matrix([[self[i][j] for j in range(index, len(self[0]))] for i in range(len(self))], self.field)

    def duplicate(self) -> Matrix:
        mat: list[list[Any]] = []
        for i, row in enumerate(self):
            mat.append([])
            for v in row:
                mat[i].append(v)
        return Matrix(mat, self.field)

    def gaussian_elimination_with(self, solve_with: Union[Vector, Matrix] = None) -> Tuple[Matrix, Union[None, Matrix]]:
        """will perform gaussian elimination on the matrix and return the result in a new matrix and the solution if solve_with is given

        Args:
            solve_with (Union[Vector, Matrix], optional): the vector/matrix to solve with. Defaults to None.

        Raises:
            TypeError: will rise if solve_with is not a Vector or Matrix

        Returns:
            Tuple[Matrix, Union[None, Matrix]]: the result of the gaussian elimination and the solution if solve_with is given
        """
        SPLIT_INDEX = len(self[0])
        res = self.duplicate()
        if solve_with is not None:
            if not (isoneof(solve_with, [Vector, Matrix]) or solve_with is None):
                raise TypeError("can only solve with a Vector or Matrix")
            res = self.concat(solve_with)
        # reorder rows
        res.reorgenize_rows()
        # performe gaussian elimination
        for curr_row_index in range(len(res)):
            # find lead value for current row
            for lead_index in range(SPLIT_INDEX):
                if res[curr_row_index][lead_index] != 0:
                    break
            else:
                continue
            # else:
            #     break
            lead_value = res[curr_row_index][lead_index]
            # make lead value equal one and change row acordingly
            if lead_value != 1:
                res.apply_operation(MOT.ROW_MULTIPLICATION,
                                    curr_row_index, 1/lead_value)
            for next_row_index in range(len(res)):
                # skip current row
                if next_row_index == curr_row_index:
                    continue
                # find if need to operate
                for next_row_lead_index in range(lead_index, SPLIT_INDEX):
                    if res[next_row_index][next_row_lead_index] != 0:
                        break
                else:
                    continue
                if next_row_lead_index == lead_index:
                    row_multiplyer = res[next_row_index][next_row_lead_index]
                    res.apply_operation(
                        MOT.ROW_MULTIPLICATION, next_row_index, -1/row_multiplyer)
                    res.apply_operation(
                        MOT.ROW_ADDITION, next_row_index, curr_row_index)
                    res.apply_operation(
                        MOT.ROW_MULTIPLICATION, next_row_index, -row_multiplyer)
        res.reorgenize_rows()
        return res.split(SPLIT_INDEX)

    def gaussian_elimination(self) -> Matrix:
        """will perform gaussian elimination on the matrix and return the result in a new matrix

        Returns:
            Matrix: the result of the gaussian elimination
        """
        return self.gaussian_elimination_with(None)[0]

    def vectorize(self) -> Vector:
        """will vectorize the matrix columns first and return the result in a new vector

        Returns:
            Vector: the result of the vectorization
        """
        arr = []
        for col in range(len(self[0])):
            arr += [self[row][col] for row in range(len(self))]
        return Vector(arr)

    @validate(None, Vector)
    def solve(self, vec: Vector) -> Union[None, Vector, Span]:
        """will solve the matrix with the given vector and return the solution

        Args:
            vec (Vector): the vector to solve with

        Raises:
            TypeError: will rise if vec is not a Vector
            NotImplementedError: will rise if there is a span of solutions and vector!=V0 ,WILL BE IMPLEMENTED LATER

        Returns:
            None: if there is no solution
            Vector: if there is a single solution
            Span: if there are multiple solutions
        """
        if not isinstance(vec, Vector):
            raise TypeError("Matrix must be solved for a vector")
        result_matrix, sol = self.gaussian_elimination_with(vec)
        sol = sol.vectorize()

        # if rank of matrix is equal to number of rows and cols return solution vector
        if result_matrix.rank == len(result_matrix) == len(result_matrix[0]):
            return sol

        # if there is a row in result matrix which has all zeros and solution vector doesnt has a zero at that row there is no solution
        for i, row in enumerate(result_matrix):
            # if (not Vector(row).has_no_zero) and sol[i] != 0:
            #     return None
            has_no_zero = check_foreach(row, lambda x: x != 0)
            if has_no_zero and sol[i] != 0:
                return None
        # otherwise there is a span of solutions

        # solve for V0
        def get_solutions_from_columns(m: Matrix) -> list[Vector]:
            def sumrows(m: Matrix) -> list[int]:
                sumrow = [0 for _ in range(len(m[0]))]
                for i in range(m.rank):
                    row = m[i]
                    for j, v in enumerate(row):
                        if v != m.field.zero:
                            sumrow[j] += 1
                return sumrow
            res = []
            for i2, v in enumerate(sumrows(m)):
                if v == m.field.zero:
                    res.append(Vector.e(i2, len(m), m.field))
            return res

        def get_solutions_from_rows(m: Matrix) -> list[Vector]:
            res = []
            # key is the index of the free factor, thve value is a dictionary which keys' are the row in which the free factor appearrs and the value are the indecies of the leading factors that depend on it
            value_depends_on_key: dict[int, dict[int, set[int]]] = dict()
            for row_index in range(m.rank):
                row = m[row_index]
                for candidate_index, candidate in enumerate(row):
                    if candidate != m.field.zero:
                        for validator_index in range(candidate_index+1, len(row)):
                            if row[validator_index] != m.field.zero:
                                if validator_index not in value_depends_on_key:
                                    value_depends_on_key[validator_index] = dict(
                                    )
                                if row_index not in value_depends_on_key[validator_index]:
                                    value_depends_on_key[validator_index][row_index] = set(
                                    )
                                value_depends_on_key[validator_index][row_index].add(
                                    candidate_index)
                        break
            for key in value_depends_on_key:
                tmp = [0 for _ in range(len(m[0]))]
                tmp[key] = m.field.one
                for row_index in value_depends_on_key[key]:
                    for index in value_depends_on_key[key][row_index]:
                        tmp[index] = -m[row_index][key]
                res.append(Vector(tmp))
            return res
        solution_span_as_arr = get_solutions_from_columns(
            result_matrix)+get_solutions_from_rows(result_matrix)
        if len(solution_span_as_arr) == 0:
            return sol
        return Span(solution_span_as_arr, sol)

    def get_eigen_vectors_of(self, eigenvalue) -> list[Vector]:
        # assume that the matrix is square
        m: Matrix = self-eigenvalue*Matrix.identity(len(self))
        m **= self.algebraic_multiplicity(eigenvalue)
        return m.kernel

    def algebraic_multiplicity(self, eigenvalue) -> int:
        return self.eigenvalues.count(eigenvalue)

    def geometric_multiplicity(self, eigenvalue) -> int:
        m: Matrix = self-eigenvalue*Matrix.identity(len(self))
        res = self.kernel
        if isinstance(res, Vector):
            res = Span([res])
        return res.dim




from ..helpers import split_many, areinstances, split_if_any
from .math_symbols import python_la_superscript_n, subscript_n


def python_la_is_number(string: str) -> bool:
    for s in string:
        if s not in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]:
            return False
    return True


def __get_superscript(string: str) -> str:
    res = ""
    j = string.index("^")
    num = string[j+1:]
    if python_la_is_number(num):
        res += string[:j]
        res += python_la_superscript_n(int(num))
    else:
        res += string
    return res


def __get_subscript(string: str) -> str:
    res = ""
    j = string.index("_")
    i = len(string)
    if "^" in string:
        i = string.index("^")
    num = string[j+1:i]
    if python_la_is_number(num):
        res += string[:j]
        res += subscript_n(int(num))
    else:
        res += string
    return res


def reorder_multiplication(string: str) -> str:

    mul = "*"
    div = "/"
    if not any([op in string for op in [mul, div]]):
        return string
    res = ""
    splits, operators = split_if_any(string, [mul, div])
    for i, op in enumerate(operators):
        if op == mul:
            res += splits[i+1]+op+splits[i]
        else:
            res += "1/"+splits[i+1]+"*"+splits[i]
    return res


def reorder_str(string: str) -> str:
    res = ""
    addition = ["+", "-"]
    splits, order = split_if_any(string, addition)
    splits = [reorder_multiplication(s) for s in splits]
    if len(order) == 0:
        return splits[0]
    for i, op in enumerate(order):
        res += splits[i]+" "+op+" "
    res += splits[-1]
    return res


def pretty_str(string: str) -> str:
    string = reorder_str(string)
    res = ""
    splits, addition_order = split_if_any(string, ["+", "-"])
    for index, s in enumerate(splits):
        i = 0
        breakers = ["^", "_"]
        while i < len(s):
            c = s[i]
            if c in breakers:
                j = 1
                while i+j < len(s) and s[i+j] not in breakers:
                    j += 1
                if c == "^":
                    res += python_la_superscript_n(s[i+1:i+j])
                elif c == "_":
                    res += subscript_n(s[i+1:i+j])
                i += j - 1
            else:
                res += c
            i += 1
        if index != len(splits)-1:
            res += " " + addition_order[index] + " "
    return res


def pretty_print(*args: list[str], **kwargs) -> None:
    if not areinstances([*args], str):
        raise TypeError("s must be a string")
    result = []
    for string in args:
        result.append(pretty_str(string))
    print(*result, **kwargs)



from __future__ import annotations
import random
from typing import Any, Union, Sequence
from .Vector import Vector
from .Field import Field
from .VectorSpace import VectorSpace
from ..utils import are_operators_implemnted, concat_horizontally, alloneof
from danielutils import validate, isoneof, areoneof
from ..BaseClasses import Span____


class Span__(Span____):
    @validate(None, [Sequence, lambda seq: areoneof(seq, [Vector])], Vector)
    def __init__(self, objects: list[Vector], offset: Vector = None) -> None:
        """initializes the span with the given vectors

        Args:
            vectors (list[Vector], optional): list of vectors to initialize the span with

        Raises:
            TypeError: if the values are not of type type
            AttributeError: if the objects don't have a field attribute
            ValueError: if the objects are not in the same field
            AttributeError: if the objects don't have all nescesary operator implemented
        """
        example_item = objects[0]
        for v in objects:
            if not v.field == example_item.field:
                raise ValueError(
                    "Span must be initialized with vectors in the same field")
        # # if vectors != []:
        # T = type(example_item)
        # for val in objects:
        #     if not isinstance(val, T):
        #         raise TypeError(
        #             "All elements of the base must be of the same type")

        # if not hasattr(example_item, "field"):
        #     raise AttributeError(
        #         "All objects must have a field attribute which is an Instance of class Field")
        # if all vectors in base are of the same field add then otherwise throw an error
        self.field: Field = example_item.field
        # for vector in objects:
        #     if vector.field != self.field:
        #         raise ValueError(
        #             "Span can only be created from vectors of the same field")
        self.vectors = objects
        if not are_operators_implemnted(type(self.vectors[0])):
            raise AttributeError(
                "Not all required operators are implemented for the class of the objects")

    @validate(None, bool)
    def __str__(self, raw: bool = False) -> str:
        """returns a string representation of the span

        Args:
            raw (bool, optional): whether to display raw data or to make it pretty. Defaults to False.

        Returns:
            str: string representation of the span
        """
        if raw:
            res = ""
            for v in self:
                res += v.__str__(raw)+'\n'
            return res
        return concat_horizontally(self.vectors, "\t")

    @validate(None, [[Vector, Span____], None, None])
    def __add__(self, other: Union[Vector, Span]) -> Span:
        """returns the span of the vectors in the span plus the given vector or span

        Args:
            other (Union[Vector, Span]): vector or span to add to self

        Raises:
            TypeError: if the other object is not of type Vector or Span
            ValueError: if the other dosen't have the smae field as self

        Returns:
            Span: span of the vectors in the span plus the given vector or span
        """
        if isinstance(other, Vector):
            other = Span([other])
        if not self.field == other.field:
            raise ValueError("Spans must be in the same field")
        return Span(list(set(self.vectors + other.vectors)))

    # def __sub__(self, other: Union[Vector, Span]) -> Span:
    #     """returns the span of the vectors in the span minus the given vector or span
    #         will perform what is mathematically equivelent to: self \ {other}
    #     Args:
    #         other (Union[Vector, Span]): vector or span to subtract from self

    #     Raises:
    #         TypeError: if the other object is not of type Vector or Span

    #     Returns:
    #         Span: span of the vectors in the span minus the given vector or span
    #     """
    #     if not isoneof(other, [Vector, Span]):
    #         raise TypeError("Span can only be subtracted by Vector|Span")
    #     if isinstance(other, Vector):
    #         other = Span([other])
    #     res = set(self.vectors) - set(other.vectors)
    #     return Span(list(res))

    @validate(None, int)
    def __getitem__(self, index: int) -> Vector:
        """returns the vector at the given index

        Args:
            index (int): index of the vector to return

        Raises:
            ValueError: if the index is out of range
        Returns:
            Vector: vector at the given index
        """
        if not (0 <= index < len(self)):
            raise ValueError("index out of range")
        return self.vectors[index]

    def __iter__(self):
        """iterates over the vectors in the span

        Returns:
            a list iterator for the vectors in the span
        """
        return iter(self.vectors)

    def __len__(self) -> int:
        """returns the number of vectors in the span

        Returns:
            int: number of vectors in the span
        """
        return len(self.vectors)

    @validate(None, Vector)
    def __contains__(self, vector: Vector) -> bool:
        """checks whether a vector is one of the original elements creating the span

        Args:
            vector (Vector): vector to check

        Raises:
            TypeError: if the vector is not of type Vector

        Returns:
            bool: True if the vector is in the span, False otherwise
        """
        if not isinstance(vector, Vector):
            raise TypeError(
                "can only check containment of objects of type 'Vector'")
        return vector in self.vectors

    def __hash__(self) -> int:
        """returns a hash value for the span

        Returns:
            int: hash value for the span
        """
        return hash((v for v in self))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Span):
            return self.vectors == other.vectors
        return False


class Span(Span__):
    @staticmethod
    @validate(Span__, Span__)
    def are_same_span(s1: Span, s2: Span) -> bool:
        """checks whether two spans are the same

        Args:
            s1 (Span): first span to check
            s2 (Span): second span to check

        Returns:
            bool: True if the two spans are the same, False otherwise
        """
        if s1.field != s2.field:
            return False
        s1 = s1.basis
        s2 = s2.basis

        max_dim = s1.field.degree  # ==s2.field.degree
        # dimention constraints
        if s1.dim == s2.dim == max_dim:
            return True
        elif (s1.dim == max_dim and s2.dim < max_dim) or (s1.dim < max_dim and s2.dim == max_dim):
            return False

        for v in s1:
            if not s2.contains(v):
                return False
        for v in s2:
            if not s1.contains(v):
                return False
        return True

    @staticmethod
    @validate(Field)
    def span_field(field: Field) -> Span:
        """returns a span of the field

        Args:
            field (Field): field to get the span of

        Raises:
            TypeError: if the field is not of type Field

        Returns:
            Span: span of the field
        """
        return Span(VectorSpace(field).standard_basis())

    @property
    def basis(self) -> Span:
        """returns the basis of the span

        Returns:
            Span: basis of the span
        """
        from .Matrix import Matrix
        result_indecies = []
        for row in Matrix.from_vectors(self.vectors).gaussian_elimination():
            for vec_index, value in enumerate(row):
                if value != self.field.zero:
                    result_indecies.append(vec_index)
                    break
        return Span([self.vectors[i] for i in result_indecies])

    @property
    def dim(self) -> int:
        """returns the dimension of the basis of the span

        Returns:
            int: dimension of the basis of the span
        """
        return len(self.basis)

    @property
    def has_lineary_dependency(self) -> bool:
        """returns whether the span has a lineary dependent vector inside it

        Returns:
            bool: True if the span has a lineary dependent vector inside it, False otherwise
        """
        # FIXME: this is not correct, or is it?
        return self != self.basis

    # FIXME they need inner product
    # def is_orthogonal(self) -> bool:
    #     pass

    # def is_orthonormal(self) -> bool:
    #     return self.toOrthonormal() == self

    @validate(None, Vector)
    def remove(self, vec: Vector) -> None:
        """removes a vector from the span

        Args:
            vec (Vector): vector to remove

        Raises:
            TypeError: if the vector is not of type Vector

        Returns:
            bool: True if the vector was removed, False otherwise
        """
        if vec in self.vectors:
            self.vectors.remove(vec)

    @validate(None, int)
    def remove_at(self, index: int) -> None:
        """removes the vector at the given index from the span

        Args:
            index (int): index of the vector to remove

        Raises:
            TypeError: if the index is not of type int
            ValueError: if the index is out of range
        """
        if not (0 <= index < len(self)):
            raise ValueError("index out of range")
        self.vectors.pop(index)

    @validate(None, Vector)
    def contains(self, vector: Vector) -> bool:
        """checks whether there is a linear combinations of vectors in the span that equals the vector

        Args:
            vector (Vector): vector to check

        Raises:
            TypeError: if the vector is not of type Vector

        Returns:
            bool: True if the vector is in the span, False otherwise
        """
        from .Matrix import Matrix
        res = False
        try:
            res = Matrix.from_vectors(self.vectors).solve(vector) != None
        except NotImplementedError:
            res = True
        return res

    def to_orthonormal(self) -> Span:
        """returns a span that is orthonormal representation of self in the same order

        Returns:
            Span: orthonormal representation of self
        """
        result = [self[0].normalize()]
        from ..la2 import StandardInnerProduct as sip
        for i in range(1, len(self.vectors)):
            current = self[i]
            curr_tag = Vector([0 for _ in range(len(self[0]))])
            for prev in result:
                curr_tag = curr_tag+sip(prev, current) * prev
            current = current-curr_tag
            result.append(current.normalize())
        return Span(result)

    @validate(None, Vector)
    def projection_of(self, v: Vector) -> Vector:
        """returns the projection of a vector onto the span

        Args:
            v (Vector): vector to project

        Raises:
            TypeError: if the vector is not of type Vector

        Returns:
            Vector: projection of the vector onto the span
        """
        res: Vector = Vector.from_size(len(v), 0)
        for w in self.to_orthonormal():
            res += v.projection_onto(w)
        return res

    @validate(None, [[int, float], None, None], [[int, float], None, None])
    def random(self, min: Union[int, float] = -10, max: Union[int, float] = 10) -> Vector:
        """Creates a random vector from the span

        Args:
            min (int, optional): the minimum value for the scalars in the linear dependecy. Defaults to -10.
            max (int, optional): the maximum value for the scalars in the linear dependecy. Defaults to 10.
        Raises:
            TypeError: if min and/or max are not [int, float]
        Returns:
            Vector: a random vector from the span
        """
        res = Vector.from_size(len(self[0]), 0)
        for v in self:
            res += random.uniform(min, max) * v
        return res



from __future__ import annotations
from typing import Any, Tuple, Sequence, Sequence, Sequence, Sequence
from ...la1 import Complex
from .PolynomialSimple import PolynomialSimple
from danielutils import validate, isoneof, NotImplemented


class PolynomialFraction____:
    pass


class PolynomialFraction__(PolynomialFraction____):
    @validate(None, Sequence, Sequence, Sequence, Sequence)
    def __init__(self, nominator_prefixes: list, nominator_powers: list, denominator_prefixes: list = [1], denominator_powers: list = [0]) -> None:
        self.nominator = PolynomialSimple(nominator_prefixes, nominator_powers)
        self.denominator = PolynomialSimple(
            denominator_prefixes, denominator_powers)

    @validate(None, [int, float, Complex, PolynomialSimple, PolynomialFraction____])
    def __add__(self, other) -> PolynomialFraction:
        if isinstance(other, PolynomialFraction):
            n1, d1 = self.nominator, self.denominator
            n2, d2 = other.nominator, other.denominator
            return PolynomialFraction.from_polynomials(n1*d2+n2*d1, d1*d2)
        elif isinstance(other, PolynomialSimple):
            return self+PolynomialFraction(other.prefixes, other.powers)
        elif isoneof(other, [int, float, Complex]):
            return self+PolynomialFraction.from_primitive(other)

    def __radd__(self, other) -> PolynomialFraction:
        return self.__add__(other)

    def __sub__(self, other) -> PolynomialFraction:
        return self + (-other)

    def __rsub__(self, other) -> PolynomialFraction:
        return other + (-self)

    def __neg__(self) -> PolynomialFraction:
        return self.__mul__(-1)

    @validate(None, [int, float, Complex, PolynomialSimple, PolynomialFraction____])
    def __mul__(self, other) -> PolynomialFraction:
        if isinstance(other, PolynomialFraction):
            return PolynomialFraction.from_polynomials(self.nominator*other.nominator, self.denominator*other.denominator)
        elif isinstance(other, PolynomialSimple):
            return self*PolynomialFraction(other.prefixes, other.powers)
        elif isoneof(other, [int, float, Complex]):
            return self*PolynomialFraction.from_primitive(other)

    def __rmul__(self, other) -> PolynomialFraction:
        return self.__mul__(other)

    @validate(None, [int, float, Complex, PolynomialSimple, PolynomialFraction____])
    def __truediv__(self, other) -> PolynomialFraction:
        if isinstance(other, PolynomialFraction):
            n1, d1 = self.nominator, self.denominator
            n2, d2 = other.nominator, other.denominator
            return PolynomialFraction.from_polynomials(n1*d2, d1*n2)
        elif isinstance(other, PolynomialSimple):
            return self / PolynomialFraction(other.prefixes, other.powers)
        elif isoneof(other, [int, float, Complex]):
            return self / PolynomialFraction.from_primitive(other)

    def __rtruediv__(self, other) -> PolynomialFraction:
        # if isinstance(other, Expression):
        #     n1, d1 = self.destructure()
        #     n2, d2 = other.destructure()
        #     return Expression.fromPolynomials(n1*d2, d1*n2)
        if isinstance(other, PolynomialSimple):
            return self / PolynomialFraction(other.prefixes, other.powers)
        elif isoneof(other, [int, float, Complex]):
            return PolynomialFraction.from_primitive(other) / self
        raise NotImplementedError(
            "Division of Expressions with {} not implemented yet".format(type(other)))

    @validate(None, [int, float, Complex, PolynomialSimple, PolynomialFraction____])
    def __eq__(self, other) -> PolynomialFraction:
        if isinstance(other, PolynomialFraction):
            n1, d1 = self.nominator, self.denominator
            n2, d2 = other.nominator, other.denominator
            # n1/d1 == n2/d2 <==> n1d2/d1d2 == n2d1/d1d2 <==> n1d2 == n2d1
            return n1*d2 == n2*d1
        elif isinstance(other, PolynomialSimple):
            return self == PolynomialFraction(other.prefixes, other.powers)
        elif isoneof(other, [int, float, Complex]):
            return self == PolynomialFraction.from_primitive(other)
        return False

    def __ne__(self, other) -> PolynomialFraction:
        return not self.__eq__(other)

    @validate(None, int)
    def __pow__(self, other) -> PolynomialFraction:
        if isinstance(other, int):
            return PolynomialFraction.from_polynomials(self.nominator**other, self.denominator**other)

    @NotImplemented
    def __rpow__(self, other) -> PolynomialFraction:
        pass

    def __str__(self) -> str:
        if self.denominator != 1:
            return f"({str(self.nominator)})/({str(self.denominator)})"
        return str(self.nominator)

    @validate(None, [int, float, Complex, PolynomialSimple, PolynomialFraction____])
    def __call__(self, value) -> Any:
        try:
            return self.nominator(value) / self.denominator(value)
        except ArithmeticError as e:
            raise e


class PolynomialFraction(PolynomialFraction__):
    @staticmethod
    @NotImplemented
    @validate(str, str)
    def from_string(string, var="x") -> PolynomialFraction:
        pass

    @staticmethod
    @validate(PolynomialSimple, PolynomialSimple)
    def from_polynomials(nominator: PolynomialSimple, denominator: PolynomialSimple) -> PolynomialFraction:
        return PolynomialFraction(nominator.prefixes, nominator.powers, denominator.prefixes, denominator.powers)

    @staticmethod
    @validate([int, float, Complex])
    def from_primitive(value) -> PolynomialFraction:
        return PolynomialFraction([value], [0])

    def destructure(self) -> Tuple[PolynomialSimple, PolynomialSimple]:
        return self.nominator, self.denominator

    @NotImplemented
    def simplify() -> PolynomialFraction:
        pass



from __future__ import annotations
from typing import Callable, Union
from ..la1 import Complex, Vector, Field, Matrix, VectorSpace
from danielutils import validate


class BilinearForm:
    # @staticmethod
    # def isBilinearForm() -> bool:
    #     pass

    # @staticmethod
    # def fromInnerProduct(prod) -> BilinearForm:
    #     pass

    # @staticmethod
    # def createSquareBilinearForm() -> BilinearForm:
    #     pass

    # @staticmethod
    # @validate(Matrix)
    # def from_matrix(mat: Matrix) -> BilinearForm:
    #     pass

    @validate(None, Callable, VectorSpace)
    def __init__(self, func: Callable[[Vector, Vector], Union[int, float, Complex]], vector_space: VectorSpace) -> None:
        self.func = func
        self.vector_space = vector_space

    # @property
    # def kernel(self):
    #     return self.to_matrix().kernel

    # @property
    # def is_symmetrical(self) -> bool:
    #     pass

    # @property
    # def is_square(self) -> Union[bool, None]:
    #     return None

    def __call__(self, v, u):
        return self.func(v, u)

    def to_matrix(self) -> Matrix:
        basis = self.vector_space.standard_basis()
        mat = [[0 for i in range(len(basis))] for j in range(len(basis))]
        for i, v in enumerate(basis):
            for j, u in enumerate(basis):
                mat[i][j] = self.func(v, u)
        return Matrix(mat, field=self.vector_space.field)

    @validate(None, Vector, Vector)
    def are_perpendicular(self, v1: Vector, v2: Vector) -> bool:
        """
        will return true if result of the bilinear form calculation is 0
        """
        return self.func(v1, v2) == 0


# class SquareBilinearForm(BilinearForm):
#     @property
#     def is_square(self) -> bool:
#         return True

#     def __call__(self, v):
#         return self.func(v, v)

#     def to_matrix(self) -> Matrix:
#         pass



from .helpers import *
from .instance_cheking import *
from .string_manipulation import *
from .polynomial_from_string_helper_functions import *
from .mathematical_helpers import *
from .math_symbols import *
from .pretty_print import *



from __future__ import annotations
from typing import Callable, Any, Union
from .Field import Field
from ..utils import composite_function, areinstances
from .Matrix import Matrix
from .Vector import Vector
from .Complex import Complex
from .VectorSpace import VectorSpace
from danielutils import validate, isoneof
from ..BaseClasses import LinearMap____


class LinearMap__(LinearMap____):
    @validate(None, VectorSpace, VectorSpace, Callable, bool)
    def __init__(self, src: VectorSpace, dst: VectorSpace, func: Callable[[Any], Any], validate: bool = False) -> None:
        """creates a new linear transformation

        Args:
            src_field (Field): The source field which elements form it are called with this opeartor
            dst_field (Field): The Field which is the output oif this transformation
            func (Callable[[Any], Any]): the transformation function
        """
        if validate:
            if not LinearMap.is_func_linear_map(func, src, dst):
                raise ValueError("func is not a linear transformation")
        self.src = src
        self.dst = dst
        self.func = func

    @validate(None, LinearMap____)
    def __add__(self, other) -> LinearMap:
        # if isoneof(other, [int, float, complex]):
        # return LinearTransformation(self.src_field, self.dst_field, lambda x, y: self.func(x, y)+other*LinearTransformation())
        if self.src == other.src and self.dst == other.dst:
            return LinearMap(self.src, self.dst, lambda x: self(x)+other(x))
        raise ValueError(
            "cant add linear transformations on diffrent fields")

    @validate(None, LinearMap____)
    def __radd__(self, other) -> LinearMap:
        return self.__add__(other)

    @validate(None, LinearMap____)
    def __sub__(self, other) -> LinearMap:
        return self.__add__(-other)

    @validate(None, LinearMap____)
    def __rsub__(self, other) -> LinearMap:
        return other+(-self)

    def __neg__(self) -> LinearMap:
        return self.__mul__(-1)

    @validate(None, [[int, float, Complex, LinearMap____], None, None])
    def __mul__(self, other) -> LinearMap:
        if isoneof(other, [int, float, Complex]):
            return LinearMap(self.src, self.dst, lambda x: self.func(x)*other)
        if isinstance(other, LinearMap):
            return self(other)

    def __rmul__(self, other) -> LinearMap:
        return self.__mul__(other)

    @validate(None, [[int, float], None, None])
    def __pow__(self, other) -> LinearMap:
        if not other == int(other) or other < 0:
            raise ValueError(
                "only non negativ integer powers are implemented and you tried to raise the transformation to {}".format(other))
        if other == 0:
            return LinearMap(self.src, self.dst, lambda x: x)
        if other == 1:
            return self
        other = int(other)
        func = composite_function(self.func, self.func)
        for _ in range(abs(other)-2):
            func = composite_function(func, self.func)
        return LinearMap(self.src, self.dst, lambda x: func(x))

    def __eq__(self, other: LinearMap) -> bool:
        if not isinstance(other, LinearMap):
            return False
        if not (self.src == other.src and self.dst == other.dst):
            return False
        return self.to_matrix() == other.to_matrix()

    @validate(None, [[Vector, Matrix, LinearMap____], None, None])
    def __call__(self, val: Union[Vector, Matrix, LinearMap]) -> Union[Vector, Matrix, LinearMap]:
        """ apply the transformation on an object
        Args:
            v (Any): the object to apply the transformation on

        Raises:
            ValueError: will raise a ValueError if the object is not in the source field
            Exception: if the transformation function itself raises an error

        Returns:
            Any: the return value of the transformation's function
        """
        if not isoneof(val, [Vector, Matrix, LinearMap]):
            raise ValueError(
                "can only apply linear transformations on vectors and matrices")
        # LinearMap
        if isinstance(val, LinearMap):
            if not self.dst == val.src:
                raise ValueError(
                    "outer dst vector space is not equal to inner src vector space")
            return LinearMap(self.src, val.dst, lambda x: self.func(val(x)))

        # Matrix or Vector
        if isinstance(val, Vector):
            if val not in self.src:
                raise ValueError("value is not in the source field")
        else:
            if not Vector(val[i][0] for i in range(val._rows)).field == self.src:
                raise ValueError("value is not in the source field")
        try:
            return self.func(val)
        except Exception as e:
            raise e


class LinearMap(LinearMap__):
    # TODO rethink data structure
    @staticmethod
    def is_func_linear_map(func: Callable[[Any, Field], Union[Vector, Matrix]], src_field: Field, dst_field: Field) -> bool:
        COUNT = 100
        V = VectorSpace(src_field)
        for _ in range(COUNT):
            a, b = src_field.random(), src_field.random()
            v1, v2 = V.random(), V.random()
            try:
                if not func(a*v1+b*v2, dst_field) == (a*func(v1, dst_field)+b*func(v2, dst_field)):
                    pass
            except Exception as e:
                pass
            if not func(a*v1+b*v2, dst_field) == (a*func(v1, dst_field)+b*func(v2, dst_field)):
                return False
        return True

    @staticmethod
    @validate(Matrix)
    def from_matrix(m: Matrix) -> LinearMap:
        return LinearMap(m.field, type(m.field)(m.field.name), lambda x: m*x)

    @staticmethod
    def from_fields(src_field, dst_field, func: Callable[[Any, Field], Any], validate: bool = False) -> LinearMap:
        return LinearMap(VectorSpace(src_field), VectorSpace(dst_field), func, validate)

    @staticmethod
    def id(field: int) -> LinearMap:
        return LinearMap(field, field, lambda x: x)

    def to_matrix(self, basis: list[Vector] = None) -> Matrix:
        if basis is None:
            basis = self.src.standard_basis()
        vectors = []
        for v in basis:
            vectors.append(self(v))
        return Matrix.from_vectors(vectors)


class LinearTransformation(LinearMap):
    pass



class LinearMap____:
    pass


class Matrix____:
    pass


class Span____:
    pass


class Vector____:
    pass


class Complex____:
    pass


class Field____:
    pass


class VectorSpace____:
    pass


class PolynomialSimple____:
    pass



from __future__ import annotations
from typing import Any
from ..Field import Field
from ..Vector import Vector
from ...BaseClasses import VectorSpace____
from danielutils import validate


class VectorSpace__(VectorSpace____):
    @validate(None, Field)
    def __init__(self, field: Field) -> None:
        """creates a new vector space

        Args:
            field (Field): The field on which the space if over
        """
        self.field = field

    def __str__(self) -> str:
        """returns a string representation of this vector space

        Returns:
            str: a string representation of this vector space
        """
        return f"V_{str(self.field)}"

    def __eq__(self, other: Any) -> bool:
        """checks if this vector space is equal to another vector space

        Args:
            other (VectorSpace): the other vector space

        Raises:
            TypeError: if other is not an instance of VectorSpace

        Returns:
            bool: True if this vector space is equal to the other vector space
        """
        if not isinstance(other, VectorSpace):
            False
        return self.field == other.field

    def __ne__(self, other: Any) -> bool:
        """checks if this vector space is not equal to another vector space

        Args:
            other (VectorSpace): the other vector space

        Raises:
            TypeError: if other is not an instance of VectorSpace

        Returns:
            bool: True if this vector space is not equal to the other vector space
        """
        return not self.__eq__(other)

    def __contains__(self, value: Any) -> bool:
        """checks if a vector is in this vector space

        Args:
            value (Vector): the vector to check

        Raises:
            TypeError: if value is not an instance of Vector

        Returns:
            bool: True if the vector is in this vector space
        """
        if not isinstance(value, Vector):
            False
        if value.field == self.field:
            return True
        return False


class VectorSpace(VectorSpace__):

    def random(self, min: Any = -10, max: Any = 10) -> Vector:
        """returns a random vector in this vector space

        Args:
            min (Any, optional): minimum value for each element in vector. Defaults to -10.
            max (Any, optional): maximum value for each element in vector. Defaults to 10.

        Returns:
            Vector: a random vector in this vector space
        """
        return Vector([self.field.random(min, max) for _ in range(self.field.degree)], self.field)

    @validate(None, int)
    def e(self, i: int) -> Vector:
        """returns the i-th standard basis vector of this vector space

        Args:
            i (int): the index of the basis vector to return

        Raises:
            TypeError: if i is not an integer
            ValueError: if i is not in the range [0, self.field.degree)

        Returns:
            Vector: the i-th standard basis vector of this vector space
        """
        if not (0 < i <= self.field.degree):
            raise ValueError("i must be between 1 and degree")
        arr = [0 if j != i-1 else 1 for j in range(self.field.degree)]
        return Vector(arr, self.field)

    def standard_basis(self) -> list[Vector]:
        """returns the standard basis of this vector space

        Returns:
            list[Vector]: the standard basis of this vector space
        """
        n = self.field.degree
        return [self.e(i+1) for i in range(n)]



# import Complex
from typing import Any, Callable
import math
from .instance_cheking import alloneof


def are_operators_implemnted(T) -> bool:
    try:
        T.__add__
        T.__radd__
        T.__sub__
        T.__rsub__
        T.__neg__
        T.__mul__
        T.__rmul__
        T.__truediv__
        T.__rtruediv__
        T.__eq__
        T.__ne__
        T.__hash__
        return True
    except AttributeError:
        return False


def almost_equal(*args):
    THRESHOLD = 0.000000000001

    def wrapper(a, b):
        if alloneof([a, b], [int, float]):
            return math.isclose(a, b, abs_tol=THRESHOLD)
        else:  # they are Complex.Complex
            try:
                return math.isclose(a.real, b.real, abs_tol=THRESHOLD) and math.isclose(a.imag, b.imag, abs_tol=THRESHOLD)
            except Exception as e:
                assert False, "shouldnt be here"
    return all([wrapper(args[0], args[i]) for i in range(1, len(args))])


def check_foreach(arr: list, condition) -> bool:
    for v in arr:
        if not condition(v):
            return False
    return True


def composite_function(f, g):
    return lambda *args: f(g(*args))


def check_forevery(arr: list[Any], amount: int, condition: Callable[[], bool]) -> bool:
    def helper(arr: list, current: list) -> bool:
        if len(current) == amount:
            return condition(*current)
        for i, v in enumerate(arr):
            if not helper(arr[i+1:], current+[v]):
                return False
        return True
    return helper(arr, [])



from __future__ import annotations
from danielutils import validate, abstractmethod


class Calculable:
    @staticmethod
    @abstractmethod
    def from_string(s: str) -> Calculable:
        pass

    def __init__(self):
        try:
            self.__add__
            self.__radd__
            self.__mul__
            self.__rmul__
            self.__neg__
            self.__sub__
            self.__rsub__
            self.__truediv__
            self.__rtruediv__
            self.__pow__
            self.__str__
            self.__eq__
            self.__ne__
        except NotImplementedError:
            raise NotImplementedError(
                "abstract class - operators are not implemented")

    @abstractmethod
    def __add__(self, other) -> Calculable:
        pass

    @abstractmethod
    def __radd__(self, other) -> Calculable:
        pass

    @abstractmethod
    def __mul__(self, other) -> Calculable:
        pass

    @abstractmethod
    def __rmul__(self, other) -> Calculable:
        pass

    @abstractmethod
    def __neg__(self) -> Calculable:
        pass

    @abstractmethod
    def __sub__(self, other) -> Calculable:
        pass

    @abstractmethod
    def __rsub__(self, other) -> Calculable:
        pass

    @abstractmethod
    def __truediv__(self, other) -> Calculable:
        pass

    @abstractmethod
    def __rtruediv__(self, other) -> Calculable:
        pass

    @abstractmethod
    def __pow__(self, other) -> Calculable:
        pass

    @abstractmethod
    def __eq__(self, other) -> Calculable:
        pass

    @abstractmethod
    def __ne__(self, other) -> Calculable:
        pass

    @abstractmethod
    def __call__(self, value):
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    def __repr__(self) -> str:
        return str(self)





def isoneof(v, lst: list) -> bool:
    for T in lst:
        if isinstance(v, T):
            return True
    return False


def alloneof(value: list, types: list):
    for v in value:
        if not isoneof(v, types):
            return False
    return True


def areinstances(arr: list, T):
    for o in arr:
        if not isinstance(o, T):
            return False
    return True



from __future__ import annotations
from .Calculable import Calculable
from ...la1 import Complex
from typing import Any
from .PolynomialSimple import PolynomialSimple
from ...utils import isoneof, validate_brackets, open_power, split_not_between_brackets, split_if_any, python_la_superscript_n, sign


class PolynomialCompact(Calculable):
    @staticmethod
    def from_string(input, var="x") -> PolynomialCompact:
        if any(bracket in input for bracket in ["(", ")", "[", "]", "{", "}"]):
            if not validate_brackets(input):
                raise ValueError("invalid brackets")
            partials1, order1 = split_not_between_brackets(
                input, ["+", "-"])
            if len(order1) < len(partials1):
                order1.insert(0, "+")
            res = 0
            for i, sub_input in enumerate(partials1):
                partials2, order2 = split_not_between_brackets(
                    open_power(sub_input), ["*"])
                partial_polys = []
                for sub in partials2:
                    sub = sub.strip("()")
                    partial_polys.append(
                        PolynomialSimple.from_string(sub, var))
                res = sign(order1[i])*PolynomialCompact(partial_polys) + res
            return res
            # for sub in subs:
            #     res = PolynomialSimple.from_string(sub_inputs[0], var)
            #     for sub_input in sub_inputs[1:]:
            #         res *= PolynomialSimple.from_string(sub_input, var)
            #     polys.append(res)
        return PolynomialSimple.from_string(input, var)

    def __init__(self, polys: list[int, float, Complex, PolynomialSimple, PolynomialCompact], powers: list[float] = None) -> None:
        if not isinstance(polys, list):
            raise TypeError("polys must be a list of PolynomialSimple")
        for p in polys:
            if not isoneof(p, [int, float, Complex, PolynomialSimple, PolynomialCompact]):
                raise TypeError(
                    "polys must be a list of PolynomialSimple or PolynomialCompact or numbers")
        if powers == None:
            powers = [1 for _ in polys]
        for p in powers:
            if not isoneof(p, [float, int]):
                raise TypeError("p must be a float or int")
        self.powers = powers
        self.polys = polys

    def expand(self) -> PolynomialSimple:
        res = PolynomialSimple([1], [0])
        for p in self.polys:
            if isinstance(p, PolynomialCompact):
                res *= p.expand()
            else:
                res *= p
        return res

    def __str__(self) -> str:
        res = ""
        for i, p in enumerate(self.polys):
            res += f"({str(p)}){python_la_superscript_n(self.powers[i])}"
        return res

    def __add__(self, other):

        if not isoneof(other, [int, float, Complex, PolynomialCompact, PolynomialSimple]):
            raise TypeError(
                "other must be [int, float, Complex, PolynomialCompact, PolynomialSimple]")
        return self.expand()+other

    def __radd__(self, other) -> PolynomialCompact:
        return self+other

    def __mul__(self, value) -> PolynomialCompact:
        return PolynomialCompact(self.polys+[value])

    def __rmul__(self, value) -> PolynomialCompact:
        return self*value

    def __call__(self, value: Any) -> Any:
        return self.expand()(value)

    def __eq__(self, other) -> bool:
        if not isoneof(other, [int, float, Complex, PolynomialSimple, PolynomialCompact]):
            return False
        if isinstance(other, PolynomialCompact):
            return self.expand() == other.expand()
        return self.expand() == other

    def __ne__(self, other) -> bool:
        return not (self == other)



from setuptools import setup, find_packages
import codecs


def read_file(path: str) -> list[str]:
    with codecs.open(path, 'r', 'utf-8') as f:
        return [l.strip() for l in f.readlines()]


README_PATH = 'README.md'
VERSION = '0.97.0'
DESCRIPTION = 'Python linear algebra liabrary'
LONG_DESCRIPTION = '\n'.join(read_file(README_PATH))
setup(
    name="python-la",
    version=VERSION,
    author="danielnachumdev (Daniel Nachum)",
    author_email="<danielnachumdev@gmail.com>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests", "archive"]),
    install_requires=["danielutils"],
    keywords=['python', 'linear algebra', 'vector', 'matrix',
              'field', 'vector field', 'span', 'linear maps', 'bilinear form', 'inner product', 'linear transformation'],
    classifiers=[
        # "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



from typing import Callable
import functools


def abstractmethod(func: Callable):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        raise NotImplementedError(
            f"{func.__module__}.{func.__name__}(...) is an abstract method and must be implemented in a derived classes")
    # same as using functools.wraps
    # wrapper.__doc__ = func.__doc__
    return wrapper



# https://unicode-table.com/en/sets/mathematical-signs/
python_la_double_n = "ℕ"
python_la_double_q = "ℚ"
python_la_double_r = "ℝ"
python_la_double_z = "ℤ"
python_la_double_c = "ℂ"
python_la_rpab = "〉"  # right pointing angle bracket
python_la_lpab = "〈"  # left pointing angle bracket
python_la_forall = "∀"
python_la_exists = "∃"
python_la_vl = "|"
python_la_sigma = "∑"
python_la_pi = "∏"
python_la_circled_plus = "⊕"
python_la_ne = "≠"
python_la_equiv = "≡"
python_la_lamda = "λ"
python_la_superscript_left_parenthesis = "⁽"
python_la_superscript_right_parenthesis = "⁾"
python_la_superscript_plus = "⁺"
python_la_superscript_minus = "⁻"
python_la_superscript_equals = "⁼"
python_la_superscript_0 = "⁰"
python_la_superscript_1 = "¹"
python_la_superscript_2 = "²"
python_la_superscript_3 = "³"
python_la_superscript_4 = "⁴"
python_la_superscript_5 = "⁵"
python_la_superscript_6 = "⁶"
python_la_superscript_7 = "⁷"
python_la_superscript_8 = "⁸"
python_la_superscript_9 = "⁹"
python_la_superscript_digits = [python_la_superscript_0, python_la_superscript_1, python_la_superscript_2, python_la_superscript_3,
                                python_la_superscript_4, python_la_superscript_5, python_la_superscript_6, python_la_superscript_7, python_la_superscript_8, python_la_superscript_9]


def python_la_superscript_n(n) -> str: return "".join(
    [python_la_superscript_digits[int(i)] for i in str(n)])


subscript_0 = "\u2080"
subscript_1 = "\u2081"
subscript_2 = "\u2082"
subscript_3 = "\u2083"
subscript_4 = "\u2084"
subscript_5 = "\u2085"
subscript_6 = "\u2086"
subscript_7 = "\u2087"
subscript_8 = "\u2088"
subscript_9 = "\u2089"
subscript_plus = "\u208A"
subscript_minus = "\u208B"
subscript_equals = "\u208C"
subscript_left_parenthesis = "\u208D"
subscript_right_parenthesis = "\u208E"
subscript_digits = [subscript_0, subscript_1, subscript_2, subscript_3,
                    subscript_4, subscript_5, subscript_6, subscript_7, subscript_8, subscript_9]


def subscript_n(n: int) -> str:
    return "".join(
        [subscript_digits[int(i)] for i in str(n)])


python_la_square_root = "√"
python_la_cube_root = "∛"
python_la_forth_root = "∜"


def python_la_nth_root(n: int) -> str:
    if n == 1 or n == 0:
        raise ValueError("n must not be 0,1")
    if n == 2:
        return python_la_square_root
    if n == 3:
        return python_la_cube_root
    if n == 4:
        return python_la_forth_root
    return python_la_superscript_n(n)+python_la_square_root
